# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from .change_trigger import get_change_trigger
from .type_factory import TypeFactory


class ValueTypeFactory(TypeFactory):
    """ The ValueTypeFactory class is the base class for all classes that
    create attribute types that have an associated stored value.
    """

    def __init__(self, default, allow_none, getter, setter, metadata):
        """ Initialise the object.

        :param default:
            is the default value of the type.
        :param allow_none:
            ``True`` if ``None`` is a valid value.
        :param getter:
            is the optional attribute getter.
        :param setter:
            is the optional attribute setter.
        :param metadata:
            is a dictionary of additional meta-data stored with the type.
        """

        super().__init__(metadata)

        self.allow_none = allow_none
        self.getter_func = getter
        self.setter_func = setter
        self._default_func = None

        self.set_default_value(default)

    def bind(self, model, value, rebind=False):
        """ This is called when a model attribute is being bound or rebound.

        :param model:
            is the model.
        :param value:
            is the attribute's new value.
        :param rebind:
            is set if the attribute already has a value.
        """

        # This default implementation does nothing.
        pass

    def clone(self):
        """ Create a clone of this instance.

        :return:
            the cloned instance.
        """

        clone = self.copy()
        clone.name = self.name
        clone._default_func = self._default_func

        return clone

    def copy(self):
        """ Create a copy of this instance.  This implementation is suitable
        for types that don't have any additional data.

        :return:
            the copied instance.
        """

        return self.__class__(self.default_value, self.allow_none,
                self.getter_func, self.setter_func, **self.metadata)

    def get_default_value(self):
        """ Get the type's default value.  This implementation is appropriate
        for immutable types.  Mutable types should reimplement it and return a
        copy of the default.

        :return:
            the default value.
        """

        return self.default_value

    def set_default_value(self, default):
        """ Set the type's default value.

        :param default:
            is the new default value.
        """

        self.default_value = self.validate(default)

    def __call__(self, getter_func):
        """ This is used to implicitly define a typed attribute by decorating
        the getter method.  The name of the method is used as the name of the
        attribute.
        """

        self.name = getter_func.__name__

        return self.getter(getter_func)

    def validate(self, value):
        """ Validate a value according to the type's constraints.  An
        exception is raised if the value is invalid.

        This must be reimplemented by a sub-class.

        :param value:
            the value to validate.
        :return:
            the validated value (which may be normalised in some way).
        """

        raise NotImplementedError

    def default(self, default_func):
        """ This is used to decorate a method that will be invoked to provide
        the default value of a typed attribute.  The name of the method should
        be the same as the name of the attribute.
        """

        clone = self.clone()
        clone._default_func = default_func

        return clone

    def getter(self, getter_func):
        """ This is used to decorate a method that will be invoked to get the
        value of a typed attribute.  The name of the method should be the same
        as the name of the attribute.
        """

        clone = self.clone()
        clone.getter_func = getter_func

        return clone

    def setter(self, setter_func):
        """ This is used to decorate a method that will be invoked to set the
        value of a typed attribute.  The name of the method should be the same
        as the name of the attribute.
        """

        clone = self.clone()
        clone.setter_func = setter_func

        return clone

    def __get__(self, instance, owner):
        """ Reimplemented to get the real value from the instance. """

        if instance is None:
            return self

        # See if this is the first time the value has been referenced.
        shadow = '_' + self.name
        value = getattr(instance, shadow)

        if value is Uninitialized:
            # We know there is a default function.
            value = self.validate(self._default_func(instance))

            self.bind(instance, value)

            setattr(instance, shadow, value)
        else:
            # Invoke an explicit getter if there is one.
            getter_func = self.getter_func

            if getter_func is not None:
                value = self.validate(getter_func(instance))

                self.bind(instance, value)

        return value

    def __set__(self, instance, value):
        """ Reimplemented to validate the value before setting it and to
        trigger any changes.
        """

        name = self.name
        trigger = get_change_trigger(instance, name)

        value = self.validate(value)
        self.bind(instance, value, rebind=True)

        # Save the old value if the attribute is being observed.
        if trigger is not None:
            old = getattr(instance, name)

        setter_func = self.setter_func

        if setter_func is not None:
            setter_func(instance, value)

        # Save the new value in the shadow attribute, even if there was an
        # explicit setter.
        shadow = '_' + self.name
        setattr(instance, shadow, value)

        # See if the value has really changed.
        if trigger is not None and self.different(old, value):
            # Avoid a circular import.
            from .attribute_change import AttributeChange

            change = AttributeChange(model=instance, name=name, new=value,
                    old=old)
            trigger.changed.emit(change)

    @classmethod
    def different(cls, value1, value2):
        """ Compare two valid values to see if they are different.

        :param value1:
            is the first value.
        :param value2:
            is the second value.
        :return:
            ``True`` if the values are different.
        """

        return value1 != value2


# A unique object that it a placeholder for an uninitialized attribute.
Uninitialized = object()
