# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


import sys

from PyQt4.QtCore import QEvent, QObject, QPoint, Qt, QTimer
from PyQt4.QtGui import (QAbstractButton, QAbstractItemView, QApplication,
        QLayout, QTreeWidget, QTreeWidgetItem)
from PyQt4.QtTest import QTest

from ....automate import IAutomatedEditor, IAutomatedListEditor, AutomationError
from ....model import adapt, Adapter, Instance, Model, notify_observers
from ....ui import IListEditor

from ..editor_delegate import EditorDelegate
from ..editor_layout_adapter import EditorLayoutAdapter
from ..visual_items import find_visual_item


# Qt should handle this.
_EDIT_KEY = Qt.Key_Enter if sys.platform == 'darwin' else Qt.Key_F2


class ListEditorComponents(Model):
    """ The ListEditorComponents class is an internal mixin for adapters that
    access the embedded components of a list editor.
    """

    # The tree widget.
    tree_widget = Instance(QTreeWidget)

    @tree_widget.default
    def tree_widget(self):
        """ Invoked to return the tree widget. """

        return find_visual_item(self.adaptee, QTreeWidget)

    # The add button.
    add_button = Instance(QAbstractButton)

    @add_button.default
    def add_button(self):
        """ Invoked to return the add button. """

        return find_visual_item(self.adaptee, QAbstractButton, name='add')

    # The remove button.
    remove_button = Instance(QAbstractButton)

    @remove_button.default
    def remove_button(self):
        """ Invoked to return the remove button. """

        return find_visual_item(self.adaptee, QAbstractButton, name='remove')

    @classmethod
    def usable_editor(cls, item):
        """ Return True if an item can be used as an editor. """

        # The tree widget must be an immediate child otherwise the criteria
        # would be far too wide.
        if find_visual_item(item, QTreeWidget, recurse=False) is None:
            return False

        if find_visual_item(item, QAbstractButton, name='add') is None:
            return False

        if find_visual_item(item, QAbstractButton, name='remove') is None:
            return False

        return True


@adapt(QLayout, to=IListEditor)
class QLayoutIListEditorAdapter(EditorLayoutAdapter, ListEditorComponents):
    """ An adapter to implement IListEditor for a layout that embeds a
    :class:`~PyQt4.QtGui.QTreeWidget`, a :class:`~PyQt4.QtGui.QAbstractButton`
    named 'add' and a :class:`~PyQt4.QtGui.QAbstractButton` named 'remove'.
    """

    # The tree widget's event filter.
    _event_filter = Instance(QObject)

    @classmethod
    def isadaptable(cls, adaptee):
        """ Determine if this adapter can adapt a particular layout.

        :param adaptee:
            is the layout.
        :return:
            ``True`` if the layout can be adapted.
        """

        return cls.usable_editor(adaptee)

    def configure(self, properties):
        """ Configure the editor. """

        tree_widget = self.tree_widget
        tree_widget.setDragDropMode(QAbstractItemView.InternalMove)
        tree_widget.installEventFilter(self._event_filter)
        tree_widget.setRootIsDecorated(False)

        self.add_button.clicked[()].connect(self._on_add_clicked)

        remove_button = self.remove_button
        remove_button.setEnabled(False)
        remove_button.clicked[()].connect(self._on_remove_clicked)

        # Apply any properties to the tree widget.
        try:
            tree_widget.pyqtConfigure(**properties)
        except AttributeError:
            pass

    @IListEditor.columns.setter
    def columns(self, value):
        """ Invoked to set the columns. """

        # Save for the standard getter.
        self._columns = value

        tree_widget = self.tree_widget

        # If there are any labels then use them for the column headings.
        labels = []

        for col_nr, col in enumerate(value):
            self.tree_widget.setItemDelegateForColumn(col_nr,
                    EditorDelegate(col.editor_factory, col.column_type,
                            self.adaptee))

            labels.append(col.label)

        tree_widget.setHeaderLabels(labels)

    @IListEditor.label_policy.getter
    def label_policy(self):
        """ Invoked to get the label policy. """

        return 'default' if self.columns[0].bind_to != '' else 'embedded'

    @IListEditor.read_only.getter
    def read_only(self):
        """ Invoked to get the read-only state. """

        return self.tree_widget.editTriggers() == QAbstractItemView.NoEditTriggers

    @read_only.setter
    def read_only(self, value):
        """ Invoked to set the read-only state. """

        tree_widget = self.tree_widget

        tree_widget.setEditTriggers(
                QAbstractItemView.NoEditTriggers if value else
                        QAbstractItemView.DoubleClicked |
                        QAbstractItemView.EditKeyPressed |
                        QAbstractItemView.AnyKeyPressed)

        tree_widget.setDragDropMode(
                QAbstractItemView.NoDragDrop if value else
                        QAbstractItemView.InternalMove)

        self.remove_button.parent().setVisible(not value)

    @IListEditor.value.getter
    def value(self):
        """ Invoked to get the editor's value. """

        tree_widget = self.tree_widget

        value = []

        for row_nr in range(tree_widget.topLevelItemCount()):
            row = self.item_factory()
            itm = tree_widget.topLevelItem(row_nr)

            for col_nr, col in enumerate(self.columns):
                data = itm.data(col_nr, Qt.EditRole)

                # The data may be ``None`` if there is an open delegate.
                if data is not None:
                    if col.bind_to != '':
                        setattr(row, col.bind_to, data)
                    else:
                        row = data

            value.append(row)

        return value

    @value.setter
    def value(self, value):
        """ Invoked to set the editor's value. """

        tree_widget = self.tree_widget

        # Save for the notifier.
        self.qt4_old_value = list(value)

        old_len = tree_widget.topLevelItemCount()
        new_len = len(value)

        # Update any rows that will be kept.
        keep = min(old_len, new_len)

        for row_nr in range(keep):
            row = value[row_nr]
            itm = tree_widget.topLevelItem(row_nr)

            for col_nr, col in enumerate(self.columns):
                if col.bind_to != '':
                    data = getattr(row, col.bind_to)
                else:
                    data = row

                if itm.data(col_nr, Qt.EditRole) != data:
                    itm.setData(col_nr, Qt.EditRole, data)
                    itm.setData(col_nr, Qt.UserRole, '')

        row_nr = keep

        if old_len < new_len:
            # Add the rows for the new elements.
            while row_nr < new_len:
                self._new_item(value[row_nr])
                row_nr += 1

        elif old_len > new_len:
            # Remove the rows for the old elements.
            while row_nr < old_len:
                tree_widget.takeTopLevelItem(row_nr)
                row_nr += 1

        self.remove_button.setEnabled(new_len != 0)

    @IListEditor.invalid_reason.getter
    def invalid_reason(self):
        """ Invoked to get the invalid reason. """

        tree_widget = self.tree_widget

        # Find an invalid item giving preference to the current item.
        itm = tree_widget.currentItem()

        if itm is None:
            reason = ''
        else:
            bad_col = tree_widget.currentColumn()
            reason = itm.data(bad_col, Qt.UserRole)

        if reason == '':
            for row_nr in range(tree_widget.topLevelItemCount()):
                itm = tree_widget.topLevelItem(row_nr)

                for col_nr in range(tree_widget.columnCount()):
                    reason = itm.data(col_nr, Qt.UserRole)
                    if reason != '':
                        bad_col = col_nr
                        break

                if reason != '':
                    break

        if reason != '':
            row_nr = tree_widget.indexOfTopLevelItem(itm)

            if len(self.columns) == 1:
                reason = "in row {0}, {1}".format(row_nr + 1, reason)
            else:
                reason = "in row {0} of '{1}', {2}".format(row_nr + 1,
                        self.columns[bad_col].label, reason)

        return reason

    def item_invalid_reason(self, invalid_reason):
        """ Update the invalid reason for the current item in the collection.

        :param index:
            is the invalid reason.  It will be an empty string if the item is
            valid.
        """

        tree_widget = self.tree_widget

        tree_widget.currentItem().setData(tree_widget.currentColumn(),
                Qt.UserRole, invalid_reason)

    @_event_filter.default
    def _event_filter(self):
        """ Invoked to return the default event filter. """

        return EventFilter(self)

    def _on_add_clicked(self):
        """ Invoked to add a new item to the collection. """

        tree_widget = self.tree_widget

        # Add a new item.
        itm = self._new_item(self.item_factory())

        self.remove_button.setEnabled(True)

        # Start editing it.
        tree_widget.setCurrentItem(itm)
        tree_widget.editItem(itm)

    def _on_remove_clicked(self):
        """ Invoked to remove the current item from the collection. """

        tree_widget = self.tree_widget

        last_row = (tree_widget.topLevelItemCount() == 1)

        self._event_filter.setRemovingRows(True)
        tree_widget.takeTopLevelItem(
                tree_widget.indexOfTopLevelItem(tree_widget.currentItem()))
        self._event_filter.setRemovingRows(False)

        if last_row:
            self.remove_button.setEnabled(False)

        self.qt4_notify(self.value)

    def _new_item(self, value):
        """ Add a new item. """

        itm = QTreeWidgetItem()

        if not self.read_only:
            itm.setFlags(itm.flags() | Qt.ItemIsEditable)

        for col_nr, col in enumerate(self.columns):
            if col.bind_to != '':
                data = getattr(value, col.bind_to)
            else:
                data = value

            itm.setData(col_nr, Qt.EditRole, data)
            itm.setData(col_nr, Qt.UserRole, '')

        self.tree_widget.addTopLevelItem(itm)

        return itm

    def qt4_order_changed(self):
        """ Invoked when the order of items changes. """

        self.qt4_notify(self.value)


class EventFilter(QObject):
    """ EventFilter is an internal class that filters a tree widget's events
    in order to detect internal moves.
    """

    def __init__(self, widget_adapter):
        """ Initialise the object. """

        super().__init__(widget_adapter.adaptee)

        self._removing_rows = False
        self._widget_adapter = widget_adapter

    def eventFilter(self, obj, event):
        """ Reimplemented to detect internal moves. """

        # We can't simply look for drop events (they don't actually occur).
        if not self._removing_rows and event.type() == QEvent.ChildRemoved:
            # Notify after the event has been fully handled.
            QTimer.singleShot(0, self._widget_adapter.qt4_order_changed)

        return super().eventFilter(obj, event)

    def setRemovingRows(self, removing):
        """ Remember if we are removing rows programmatically. """

        self._removing_rows = removing


@adapt(QLayout, to=IAutomatedListEditor)
class QLayoutIAutomatedListEditorAdapter(Adapter, ListEditorComponents):
    """ An adapter to implement IAutomatedListEditor for a layout that embeds a
    :class:`~PyQt4.QtGui.QTreeWidget`, a :class:`~PyQt4.QtGui.QAbstractButton`
    named 'add' and a :class:`~PyQt4.QtGui.QAbstractButton` named 'remove'.
    """

    @classmethod
    def isadaptable(cls, adaptee):
        """ Determine if this adapter can adapt a particular layout.

        :param adaptee:
            is the layout.
        :return:
            ``True`` if the layout can be adapted.
        """

        return cls.usable_editor(adaptee)

    def simulate_append(self, id, delay, value):
        """ Simulate the user appending a value to the list.

        :param id:
            is the (possibly scoped) identifier of the widget.  This is
            normally used in exceptions.
        :param delay:
            is the delay in milliseconds between simulated events.
        :param value:
            is the value to append.
        """

        columns = IListEditor(self.adaptee).columns

        nr_rows = self.tree_widget.topLevelItemCount()

        col0_name = columns[0].bind_to
        col0_value = getattr(value, col0_name) if col0_name != '' else value

        QTest.mouseClick(self.add_button, Qt.LeftButton, Qt.NoModifier,
                QPoint(), delay)
        self._update_editor('append', id, delay, col0_value)

        for col_nr in range(1, len(columns)):
            self._open_editor('update', id, delay,
                    getattr(value, columns[col_nr].bind_to),
                    nr_rows, col_nr)

        # Give Qt some time to tidy up and update the model.
        QTest.qWait(100)

    def simulate_move(self, id, delay, from_index, to_index):
        """ Simulate the user moving a value within the list.

        :param id:
            is the (possibly scoped) identifier of the widget.  This is
            normally used in exceptions.
        :param delay:
            is the delay in milliseconds between simulated events.
        :param from_index:
            is the initial index of the value being moved.
        :param to_index:
            is the final index of the value being moved.
        """

        # QTest doesn't support simulating DND properly (if at all).  Therefore
        # we just manipulate the underlying widget.

        ilistedit = IListEditor(self.adaptee)
        old = ilistedit.value

        tree_widget = self.tree_widget

        self._validate_index('move', id, to_index)
        self._set_row('move', id, delay, from_index)
        itm = tree_widget.takeTopLevelItem(from_index)
        tree_widget.insertTopLevelItem(to_index, itm)
        self._set_row('move', id, delay, to_index)

        new = ilistedit.value
        notify_observers('value', ilistedit, new, old)

    def simulate_remove(self, id, delay, index):
        """ Simulate the user removing a value from the list.

        :param id:
            is the (possibly scoped) identifier of the widget.  This is
            normally used in exceptions.
        :param delay:
            is the delay in milliseconds between simulated events.
        :param index:
            is the index of the value to remove.
        """

        self._set_row('remove', id, delay, index)
        QTest.mouseClick(self.remove_button, Qt.LeftButton, Qt.NoModifier,
                QPoint(), delay)

    def simulate_update(self, id, delay, value, index, column=None):
        """ Simulate the user updating a value within the list.

        :param id:
            is the (possibly scoped) identifier of the widget.
        :param delay:
            is the delay in milliseconds between simulated events.
        :param value:
            is the new value.
        :param index:
            is the index of the value to update.
        :param column:
            is the name of the attribute within the model to update.  It should
            not be specified if the list's elements are a simple type.
        """

        columns = IListEditor(self.adaptee).columns

        # Get the column number.
        if column is not None:
            for col_nr, col in enumerate(columns):
                if col.bind_to == '':
                    raise AutomationError(id, "update",
                            "a column must not be specified when the elements "
                            "of the list are a simple type")

                if col.bind_to == column:
                    break
            else:
                raise AutomationError(id, 'update',
                        "no such column '{0}'".format(column))
        elif columns[0].bind_to != '':
            raise AutomationError(id, "update",
                    "a column must be specified when the elements of the list "
                    "are a model")
        else:
            col_nr = None

        # Open an editor.
        self._open_editor('update', id, delay, value, index, col_nr)

        # Give Qt some time to tidy up and update the model.
        QTest.qWait(100)

    def _open_editor(self, command, id, delay, value, index, col_nr):
        """ Open an editor to set a value. """

        self._set_row(command, id, delay, index, col_nr)
        QTest.keyClick(self.tree_widget, _EDIT_KEY, Qt.NoModifier, delay)

        self._update_editor(command, id, delay, value)

    def _update_editor(self, command, id, delay, value):
        """ Wait for an editor to appear and use it to set the value. """

        # We can't assume either that there is an editor available, or that it
        # might soon appear.
        timeout = 50
        editor = self._editor_widget()
        while editor is None:
            if timeout < 0:
                raise AutomationError(id, command, "no editor widget found")

            QTest.qWait(5)
            timeout -= 5
            editor = self._editor_widget()

        IAutomatedEditor(editor).simulate_set(id, delay, value)
        QTest.keyClick(editor, Qt.Key_Return, Qt.NoModifier, delay)

    def _editor_widget(self):
        """ Return the widget that seems to implement the current editor. """

        # It's a shame that Qt doesn't provide an appropriate call.
        editor = QApplication.focusWidget()

        if editor is None:
            return None

        if not self.tree_widget.isAncestorOf(editor):
            return None

        return editor

    def _set_row(self, command, id, delay, row, col=None):
        """ Set the current row. """

        self._validate_index(command, id, row)

        if delay >= 0:
            QTest.qWait(delay)

        # This is the easiest way to make the row current and visible.
        itm = self.tree_widget.topLevelItem(row)

        if col is None:
            self.tree_widget.setCurrentItem(itm)
        else:
            self.tree_widget.setCurrentItem(itm, col)

    def _validate_index(self, command, id, index):
        """ Validate an index. """

        if index < 0 or index >= self.tree_widget.topLevelItemCount():
            raise AutomationError(
                    id, command, "invalid index: {0}".format(index))
