# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from PyQt4.QtGui import QAction, QMenu

from ....ui import IAction, IActionCollection, IMenu


class PlaceHolder(QAction):
    """ An invisible place holder action.  It also acts as a namespace for some
    related utility functions.
    """

    def __init__(self, id, parent):
        """ Initialise the place holder. """

        super().__init__(parent, objectName=id, visible=False)

        # The place isn't occupied by a real action.
        self.occupied = False

    @classmethod
    def add_action(cls, action, widget):
        """ Try and add an action to a widget.

        :param action:
            is the action.
        :param widget:
            is the widget.
        :return:
            ``True`` if the action was added.  ``False`` is returned if the
            widget couldn't accommodate the action's desired positioning.
        """

        action_id = IAction(action).id

        for act in widget.actions():
            if act.isSeparator():
                continue

            submenu = act.menu()
            if submenu is None:
                if act.objectName() == action_id:
                    if not isinstance(act, cls):
                        raise ValueError(
                                "an action '{0}' already exists".format(
                                        action_id))

                    # No need to check if it is occupied because the occupier
                    # would have been found first.
                    act.occupied = True

                    widget.insertAction(act, action)

                    return True
            else:
                if IMenu(submenu).add_action(action):
                    return True

        return False

    @classmethod
    def add_action_collection(cls, action_collection, widget):
        """ Try and add an action collection to a widget.

        :param action_collection:
            is the action collection.
        :param widget:
            is the widget.
        :return:
            ``True`` if the collection was added.  ``False`` is returned if the
            widget couldn't accommodate the collection's desired positioning.
        """

        iactioncollection = IActionCollection(action_collection)

        for act in widget.actions():
            if act.isSeparator():
                continue

            submenu = act.menu()
            if submenu is None:
                if act.objectName() == iactioncollection.id:
                    if not isinstance(act, cls) or act.occupied:
                        raise ValueError(
                                "an action collection '{0}' already "
                                "exists".format(iactioncollection.id))

                    act.occupied = True

                    # Create a sub-menu if needed.
                    if iactioncollection.text != '':
                        menu = cls.as_QMenu(action_collection, widget)
                        widget.insertMenu(act, menu)
                    else:
                        for action in reversed(iactioncollection.actions):
                            widget.insertAction(act,
                                    cls.as_QAction(action, widget))

                    return True
            else:
                if IMenu(submenu).add_action(action_collection):
                    return True

        return False

    @classmethod
    def as_QMenu(cls, action_collection, widget):
        """ Make sure an action collection is a QMenu. """

        iactioncollection = IActionCollection(action_collection)

        menu = QMenu(iactioncollection.text, widget,
                objectName=iactioncollection.id)

        for action in iactioncollection.actions:
            menu.addAction(cls.as_QAction(action, menu))

        return menu

    @classmethod
    def as_QAction(cls, action, widget):
        """ Make sure an action is a QAction. """

        if isinstance(action, QAction):
            return action

        if action == '':
            action = QAction(widget)
            action.setSeparator(True)
        else:
            action = cls(action, widget)

        return action
