# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from PyQt4.QtCore import Qt
from PyQt4.QtGui import QAbstractItemView, QListWidget, QListWidgetItem
from PyQt4.QtTest import QTest

from ....automate import AutomationError, IAutomatedOptionSelector
from ....model import adapt, Adapter
from ....ui import IDisplay, IOptionList

from ..editor_widget_adapter import EditorWidgetAdapter


@adapt(QListWidget, to=IOptionList)
class QListWidgetIOptionListAdapter(EditorWidgetAdapter):
    """ An adapter to implement IOptionList for a QListWidget. """

    def configure(self, properties):
        """ Configure the editor. """

        widget = self.adaptee

        widget.setEditTriggers(QAbstractItemView.NoEditTriggers)
        widget.itemSelectionChanged.connect(self._item_selection_changed)

        super().configure(properties)

    @IOptionList.labels.setter
    def labels(self, value):
        """ Invoked to set the labels. """

        # Save for the standard getter.
        self._labels = value

        widget = self.adaptee

        # For each item see if we have a new label allowing for the items
        # having been sorted.
        nr_labels = len(value)

        for row in range(widget.count()):
            itm = widget.item(row)

            try:
                idx = self.options.index(itm.data(Qt.UserRole))
            except ValueError:
                idx = -1

            if idx >= 0 and idx < nr_labels:
                itm.setText(value[idx])

        if self.sorted:
            widget.sortItems()

    @IOptionList.options.setter
    def options(self, value):
        """ Invoked to set the options. """

        # Save for the standard getter.
        self._options = value

        widget = self.adaptee

        blocked = widget.blockSignals(True)

        # Create or delete items as required to match the number of options.
        nr_items = widget.count()
        nr_options = len(value)

        if nr_items > nr_options:
            for i in range(nr_options, nr_items):
                widget.takeItem(nr_options)

        elif nr_items < nr_options:
            for i in range(nr_items, nr_options):
                try:
                    as_string = self.labels[i]
                except IndexError:
                    option = value[i]
                    idisplay = IDisplay(option, exception=False)
                    as_string = '' if idisplay is None else idisplay.name
                    if as_string == '':
                        as_string = str(option)

                QListWidgetItem(as_string, widget)

        # Set the options before sorting.
        for idx, option in enumerate(value):
            widget.item(idx).setData(Qt.UserRole, option)

        if self.sorted:
            widget.sortItems()

        widget.setCurrentItem(None)
        widget.blockSignals(blocked)

    @IOptionList.sorted.setter
    def sorted(self, value):
        """ Invoked to set the sorted state. """

        # Save for the standard getter.
        self._sorted - value

        if value:
            self.adaptee.sortItems()

    @IOptionList.value.getter
    def value(self):
        """ Invoked to get the editor's value. """

        itm = self.adaptee.currentItem()

        return None if itm is None else itm.data(Qt.UserRole)

    @value.setter
    def value(self, value):
        """ Invoked to set the editor's value. """

        widget = self.adaptee

        if value is None:
            itm = None
        else:
            for idx in range(widget.count()):
                itm = widget.item(idx)

                if itm.data(Qt.UserRole) == value:
                    break
            else:
                raise ValueError("'{0}' is not a valid option".format(value))

        # Save for the notifier.
        self.qt4_old_value = value

        blocked = widget.blockSignals(True)
        widget.setCurrentItem(itm)
        widget.blockSignals(blocked)

    @IOptionList.invalid_reason.getter
    def invalid_reason(self):
        """ Invoked to get the invalid reason. """

        if self.allow_none:
            return ''

        itm = self.adaptee.currentItem()

        return "nothing selected" if itm is None else ''

    @IOptionList.read_only.getter
    def read_only(self):
        """ Invoked to get the read-only state. """

        return self.adaptee.selectionMode() == QAbstractItemView.NoSelection

    @read_only.setter
    def read_only(self, value):
        """ Invoked to set the read-only state. """

        self.adaptee.setSelectionMode(
                QAbstractItemView.NoSelection if value
                        else QAbstractItemView.SingleSelection)

    def _item_selection_changed(self):
        """ Invoked when the item selection changes. """

        itm = self.adaptee.currentItem()

        self.qt4_notify(None if itm is None else itm.data(Qt.UserRole))


@adapt(QListWidget, to=IAutomatedOptionSelector)
class QListWidgetIAutomatedOptionSelectorAdapter(Adapter):
    """ An adapter to implement IAutomatedOptionSelector for the widget. """

    def simulate_select(self, id, delay, index):
        """ Simulate the user selecting an option.

        :param id:
            is the (possibly scoped) identifier of the widget.
        :param delay:
            is the delay in milliseconds between simulated events.
        :param index:
            is the index of the option.
        """

        widget = self.adaptee

        if row < 0 or row >= widget.count():
            raise AutomationError(id, command, "invalid option index")

        if delay >= 0:
            QTest.qWait(delay)

        # This is the easiest way to make the row current and visible.
        widget.setCurrentRow(row)
