# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from PyQt4.QtCore import QEvent, QObject

from ...model import Instance, observe
from ...ui import IView

from .object_adapter import ObjectAdapter


class WidgetAdapter(ObjectAdapter):
    """ The WidgetAdapter class is a base class for adapters than implement the
    remainder of the :class:`~dip.pui.IView` interface for Qt widgets.
    """

    # The close event filter.
    _close_event_filter = Instance(QObject)

    @IView.enabled.getter
    def enabled(self):
        """ Invoked to get the enabled state. """

        return self.adaptee.isEnabled()

    @enabled.setter
    def enabled(self, enabled):
        """ Invoked to set the enabled state. """

        self.adaptee.setEnabled(enabled)

    @IView.status_tip.getter
    def status_tip(self):
        """ Invoked to get the status tip. """

        return self.adaptee.statusTip()

    @status_tip.setter
    def status_tip(self, value):
        """ Invoked to set the status tip. """

        self.adaptee.setStatusTip(value)

    @IView.tool_tip.getter
    def tool_tip(self):
        """ Invoked to get the tool tip. """

        return self.adaptee.toolTip()

    @tool_tip.setter
    def tool_tip(self, value):
        """ Invoked to set the tool tip. """

        self.adaptee.setToolTip(value)

    @IView.visible.getter
    def visible(self):
        """ Invoked to get the view's visibility. """

        return self.adaptee.isVisible()

    @visible.setter
    def visible(self, value):
        """ Invoked to set the view's visibility. """

        self.adaptee.setVisible(value)

    @IView.whats_this.getter
    def whats_this(self):
        """ Invoked to get the "What's This?" help. """

        return self.adaptee.whatsThis()

    @whats_this.setter
    def whats_this(self, value):
        """ Invoked to set the "What's This?" help. """

        self.adaptee.setWhatsThis(value)

    @IView.window_title.getter
    def window_title(self):
        """ Invoked to get the window title. """

        return self.adaptee.windowTitle()

    @window_title.setter
    def window_title(self, value):
        """ Invoked to set the window title. """

        self.adaptee.setWindowTitle(value)

    def set_focus(self):
        """ Give the focus to the view. """

        self.adaptee.setFocus()

    @observe('close_request_handlers')
    def _close_request_handlers_changed(self, change):
        """ Invoked when the list of close request handlers changes. """

        if self._close_event_filter is not None:
            if len(change.new) == 0:
                self.adaptee.removeEventFilter(self._close_event_filter)
                self._close_event_filter = None
        elif len(change.new) != 0:
            self._close_event_filter = CloseEventFilter(self)
            self.adaptee.installEventFilter(self._close_event_filter)


class CloseEventFilter(QObject):
    """ CloseEventFilter is an internal class that filters a widget's close
    events.
    """

    def __init__(self, widget_adapter):
        """ Initialise the object. """

        super().__init__()

        self._widget_adapter = widget_adapter

    def eventFilter(self, obj, event):
        """ Reimplemented to handle any close events. """

        if event.type() == QEvent.Close:
            for handler in self._widget_adapter.close_request_handlers:
                if not handler(obj):
                    event.ignore()
                    break
            else:
                event.accept()

            return True

        return super().eventFilter(obj, event)
