# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import (Bool, Callable, get_attribute_type, get_model_types, List,
        resolve_attribute_path)
from ..toolkits import Toolkit

from .i_action import IAction
from .i_container import IContainer
from .i_controller import IController
from .i_view_binding import IViewBinding
from .view_factory import ViewFactory


class ContainerFactory(ViewFactory):
    """ The ContainerFactory class is the base factory class for all views that
    can contain sub-views.  A sub-class determines the layout of the sub-views.
    """

    # The list of action factories.  Each is called with the model as arguments
    # and must return an object that implements or can be adapted to either the
    # :class:`~dip.ui.IAction` or :class:`~dip.ui.ActionCollection` interfaces.
    actions = List(Callable())

    # This is ``True`` if the contents should be automatically populated from
    # a model's attribute types (if no contents where explicitly set).
    auto_populate = Bool(False)

    # The contents of the view.
    contents = List()

    def __init__(self, *contents, **properties):
        """ Initialise the factory.

        :param \*contents:
            is the list of the view's contents.  An individual item can either
            be a :class:`~dip.ui.ViewFactory` instance or a string.  Strings
            are are assumed to be the names of attributes within a model.  Such
            names may include one or more periods to specify an
            :term:`attribute path`.
        :param \*\*properties:
            are keyword arguments used as property names and values that are
            applied to the view.
        """

        super().__init__(**properties)

        self.contents = contents

    def create(self, model, controller, parent):
        """ Create a view instance.

        :param model:
            is the model.
        :param controller:
            is the controller.
        :param parent:
            is the optional parent view.
        :return:
            the view.
        """

        view = super().create(model, controller, parent)

        icontainer = IContainer(view)
        icontroller = IController(controller)

        # Create any actions.
        actions = []

        for action_factory in self.actions:
            action = action_factory(model)
            if isinstance(action, IAction):
                icontroller.add_action(action)

        icontainer.actions = actions

        # Create any sub-views.
        subviews = self.create_subviews(model, controller, view)
        self.configure_subviews(view, subviews)

        icontainer.views = subviews

        return view

    def create_subviews(self, model, controller, view):
        """ Create the sub-views for a containing view.

        :param model:
            is the model.
        :param controller:
            is the controller.
        :param view:
            is the containing view.
        :return:
            the sub-views.
        """

        # If there are no explicit contents then inspect any model.
        contents = self.contents

        if len(contents) == 0 and self.auto_populate:
            contents = get_model_types(type(model)).keys()

        # Go through the contents invoking the corresponding factories.
        subviews = []

        for content in contents:
            if isinstance(content, str):
                name, submodel = resolve_attribute_path(content, model)
                attribute_type = get_attribute_type(submodel, name)

                # Create an editor factory based on the type.
                editor_factory_factory = Toolkit.declarative_factory_for_attribute_type(attribute_type)
                editor_factory = editor_factory_factory(content)

                # Create the editor from the factory.
                subview = editor_factory.create(model, controller, view)
            elif isinstance(content, ViewFactory):
                subview = content.create(model, controller, view)
            else:
                subview = self.create_additional_content(content)

            subviews.append(subview)

        return subviews

    def configure_subviews(self, view, subviews):
        """ Configure a sequence of sub-views before adding them to the
        containing view.

        :param view:
            is the containing view.
        :param subviews:
            is the sequence of sub-views.
        """

        # This default implementation does nothing.
        pass

    def create_additional_content(self, content):
        """ Create additional content to be added to the list of sub-views.
        This default implementation will always raise an exception.

        :param content:
            the content from which additional content can be created.
        :return:
            the additional content.
        """

        raise TypeError("the contents of a view must be either an instance of "
                "a ViewFactory sub-class or a string, not "
                "'{0}'".format(type(content).__name__))

    def finalize(self, view):
        """ Reimplemented to finalize the sub-views.

        :param view:
            is the view created by this factory.
        """

        for subview in IContainer(view).views:
            binding = IViewBinding(subview, adapt=False, exception=False)
            if binding is not None:
                binding.factory.finalize(subview)

        super().finalize(view)
