# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Callable, Enum

from .editor_factory import EditorFactory
from .i_line_editor import ILineEditor


class LineEditor(EditorFactory):
    """ The LineEditor class implements a line editor factory for string
    attributes.
    """

    # This specifies whether or not a value is required.  It is applied after
    # any validator has determined that the value is valid.  'no' means that a
    # value is not required (i.e. it can be an empty string).  'yes' means that
    # a value is required.  'stripped' means that a value with no leading or
    # trailing whitespace is required.  If not given then any corresponding
    # value of the type's meta-data is used instead.
    required = Enum('no', 'yes', 'stripped', allow_none=True)

    # The validator factory.  It is called to create a validator and will be
    # passed a single argument which is the :term:`view` that will use the
    # validator.  If not given then any corresponding value of the type's
    # meta-data is used instead.
    validator = Callable()

    # The specific interface implemented by the view.
    view_interface = ILineEditor

    def configure_using_metadata(self, view, metadata):
        """ Configure the view possibly using values from the bound attribute's
        meta-data.

        :param view:
            is the view to configure.
        :param metadata:
            is the attribute's metadata,
        """

        ilineeditor = ILineEditor(view)

        required = self.required
        if required is None:
            required = metadata.get('required')

        if required is not None:
            ilineeditor.required = required

        validator = self.validator
        if validator is None:
            validator = metadata.get('validator')

        if validator is not None:
            ilineeditor.validator = validator

        super().configure_using_metadata(view, metadata)
