# Copyright (c) 2012 Riverbank Computing Limited.
#
# This file is part of dip.
#
# This file may be used under the terms of the GNU General Public License
# v2 or v3 as published by the Free Software Foundation which can be found in
# the files LICENSE-GPL2.txt and LICENSE-GPL3.txt included in this package.
# In addition, as a special exception, Riverbank gives you certain additional
# rights.  These rights are described in the Riverbank GPL Exception, which
# can be found in the file GPL-Exception.txt in this package.
#
# This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
# WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.


from ..model import Bool, Enum, List, Str

from .i_editor_binding import IEditorBinding
from .editor_factory import EditorFactory
from .i_option_selector import IOptionSelector


class OptionSelectorFactory(EditorFactory):
    """ The OptionSelectorFactory class is a base class for editors that allow
    an option to be selected.
    """

    # This is set if ``None`` is allowed as a valid value.  The default value
    # is taken from the bound attribute's type.
    allow_none = Bool(None, allow_none=True)

    # The optional option labels.  These are used to visualise the
    # corresponding options.
    labels = List(Str())

    # The options.  If an option has no corresponding label then it will be
    # adapted to the :class:`~dip.ui.IDisplay` interface and its
    # :attr:`~dip.ui.IDisplay.name` used (which defaults to its string
    # representation).
    options = List()

    # This is set of the options should be sorted.
    sorted = Bool(False)

    def configure(self, view):
        """ Reimplemented to configure the IOptionSelector specific attributes.

        :param view:
            is the view to configure.
        """

        attribute_type = IEditorBinding(view).attribute_type
        if isinstance(attribute_type, Enum):
            options = attribute_type.members
            labels = [self.get_natural_name(option) for option in options]
        else:
            options = self.options
            labels = self.labels

        ioptionselector = IOptionSelector(view)

        if self.allow_none is None:
            ioptionselector.allow_none = attribute_type.allow_none
        else:
            ioptionselector.allow_none = self.allow_none

        ioptionselector.sorted = self.sorted
        ioptionselector.labels = labels
        ioptionselector.options = options

        super().configure(view)
