# Copyright (c) 2012 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from PyQt4.QtCore import QBuffer, QByteArray, QPoint

from dip.io import IStorage
from dip.io.codecs.xml import XmlCodec
from dip.io.storage import QIODeviceStorage
from dip.model import (adapt, Any, Bool, Dict, Enum, Float, implements,
        Instance, Int, Interface, List, Model, Set, Str, Tuple)
from dip.ui import Application


@implements(IStorage)
class MemoryStorage(QIODeviceStorage):
    """ The test memory storage. """

    codec = XmlCodec(format='test.xml.codec')

    memory_storage = Instance(QByteArray)

    def qiodevice(self, location):
        return QBuffer(self.memory_storage)

    def explicit_location(self, location): return None
    def implicit_location(self, model): return None
    def readable_location(self, location): return None
    def writeable_location(self, location): return None

    @memory_storage.default
    def memory_storage(self):

        return QByteArray()


class SubModel(Model):
    """ A test model used as an attribute of a main model. """

    value = Int(10)


class ITest(Interface):
    """ A test interface for attributes of a main model. """

    value = Int(20)


@implements(ITest)
class TestImpl(Model):
    """ A test model that implements the test interface. """


class TestModel(unittest.TestCase):
    """ This tests the XML decoder and encoder. """

    def test_empty_model(self):
        """ Test an empty model. """

        storage = MemoryStorage()

        class Klass(Model):
            pass

        k_write = Klass()

        storage.write(k_write, location=None)
        k_read = storage.read(Klass(), location=None)

    def test_Any(self):
        """ Test an Any attribute. """

        class Klass(Model):
            attr = Any(5)

        self._test_attr(Klass)

    def test_Any_pickle(self):
        """ Test an Any attribute that will use pickle. """

        class Klass(Model):
            attr = Any(QPoint(10, 20))

        self._test_attr(Klass)

    def test_Bool(self):
        """ Test a Bool attribute. """

        class Klass(Model):
            attr = Bool(True)

        self._test_attr(Klass)

    def test_Dict(self):
        """ Test a Dict attribute. """

        class Klass(Model):
            attr = Dict(Int(), Str())

        storage = MemoryStorage()

        m_write = Klass(attr={1: 'one', 2: 'two'})

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertDictEqual(m_read.attr, m_write.attr)

    def test_Enum(self):
        """ Test an Enum attribute. """

        class Klass(Model):
            attr = Enum('one', 'two', 'three')

        self._test_attr(Klass)

    def test_Float(self):
        """ Test an Int attribute. """

        class Klass(Model):
            attr = Float(10.33)

        self._test_attr(Klass)

    def test_typed_Instance(self):
        """ Test a typed Instance attribute. """

        class Klass(Model):
            attr = Instance(SubModel)

        storage = MemoryStorage()

        m_write = Klass(attr=SubModel())

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertEqual(m_read.attr.value, m_write.attr.value)

    def test_typed_Instance_None(self):
        """ Test a typed Instance attribute with a value of None. """

        class Klass(Model):
            attr = Instance(SubModel)

        storage = MemoryStorage()

        m_write = Klass()

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertIsNone(m_read.attr)

    def test_untyped_Instance(self):
        """ Test an untyped Instance attribute. """

        class Klass(Model):
            attr = Instance()

        storage = MemoryStorage()

        m_write = Klass(attr=SubModel())

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertEqual(m_read.attr.value, m_write.attr.value)

    def test_typed_Instance_Interface(self):
        """ Test an Instance attribute typed with an interface. """

        class Klass(Model):
            attr = Instance(ITest)

        storage = MemoryStorage()

        m_write = Klass(attr=TestImpl())

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertEqual(m_read.attr.value, m_write.attr.value)

    def test_Int(self):
        """ Test an Int attribute. """

        class Klass(Model):
            attr = Int(10)

        self._test_attr(Klass)

    def test_List(self):
        """ Test a List attribute. """

        class Klass(Model):
            attr = List(Int())

        storage = MemoryStorage()

        m_write = Klass(attr=[1, 2, 3, 5])

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertListEqual(m_read.attr, m_write.attr)

    def test_untyped_List(self):
        """ Test an untyped List attribute. """

        class Klass(Model):
            attr = List()

        storage = MemoryStorage()

        m_write = Klass(attr=[1, 2.0, '3', True, None])

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertListEqual(m_read.attr, m_write.attr)

    def test_List_typed_Instance(self):
        """ Test a List attribute typed as a typed instance. """

        class Klass(Model):
            attr = List(SubModel)

        storage = MemoryStorage()

        m_write = Klass(attr=[SubModel()])

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertEqual(m_read.attr[0].value, m_write.attr[0].value)

    def test_List_typed_Instance_Interface(self):
        """ Test a List attribute typed as a typed instance interface. """

        class Klass(Model):
            attr = List(ITest)

        storage = MemoryStorage()

        m_write = Klass(attr=[TestImpl()])

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertEqual(m_read.attr[0].value, m_write.attr[0].value)

    def test_List_untyped_Instance(self):
        """ Test a List attribute typed as an untyped instance. """

        class Klass(Model):
            attr = List(Instance())

        storage = MemoryStorage()

        m_write = Klass(attr=[SubModel()])

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertEqual(m_read.attr[0].value, m_write.attr[0].value)

    def test_Set(self):
        """ Test a Set attribute. """

        class Klass(Model):
            attr = Set(Int())

        storage = MemoryStorage()

        m_write = Klass(attr=set((1, 2, 3, 5)))

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertSetEqual(m_read.attr, m_write.attr)

    def test_Str(self):
        """ Test a Str attribute. """

        class Klass(Model):
            # Note the use of characters that need to be escaped.
            attr = Str('<>@"')

        self._test_attr(Klass)

    def test_Tuple(self):
        """ Test a Tuple attribute. """

        class Klass(Model):
            attr = Tuple(Int())

        storage = MemoryStorage()

        m_write = Klass(attr=(1, 2, 3, 5))

        storage.write(m_write, location=None)
        m_read = storage.read(Klass(), location=None)

        self.assertTupleEqual(m_read.attr, m_write.attr)

    def _test_attr(self, factory):
        """ Check that the (non-collection) attr attribute of a model can be
        encoded and decoded.
        """

        m_write = factory()

        storage = MemoryStorage()

        storage.write(m_write, location=None)
        m_read = storage.read(factory(), location=None)

        self.assertEqual(m_read.attr, m_write.attr)


if __name__ == '__main__':
    unittest.main()
