# Copyright (c) 2011 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

import gc
import weakref

from dip.model import (adapt, Adapter, implements, Instance, Interface,
        isadapted, Model, Str)


class Iface(Interface):
    """ A test interface. """

    attr = Str('iface')


class Iface2(Interface):
    """ A second test interface. """

    attr2 = Str('iface2')


class SubIface(Iface):
    """ A derived test interface. """


class Klass:
    """ A test class that needs to be adapted. """


class Klass2:
    """ A second test class that needs to be adapted. """


class Klass3:
    """ A third test class that needs to be adapted. """


class Klass4:
    """ A fourth test class that needs to be adapted. """


@adapt(Klass, to=Iface)
class KlassAdapter(Adapter):
    """ A test adapter for a single interface. """


@adapt(Klass2, to=[Iface, Iface2])
class Klass2Adapter(Adapter):
    """ A test adapter for multiple interfaces. """


@adapt(Klass3, to=SubIface)
class Klass3Adapter(Adapter):
    """ A test adapter for derived interfaces. """


@adapt(Klass4, to=Iface)
class Klass4Adapter(Adapter):
    """ A test adapter that will not adapt any object. """

    @classmethod
    def isadaptable(cls, obj):
        return False


class TestAdapter(unittest.TestCase):
    """ This tests adapters. """

    def test_nonadapter(self):
        """ Check adapt rejects a non-adapter. """

        with self.assertRaises(TypeError):
            @adapt(Klass, to=Iface)
            class NonAdapter: pass

    def test_nontype(self):
        """ Check adapt rejects a non-type. """

        with self.assertRaises(TypeError):
            @adapt('bad', to=Iface)
            class NonType: pass

    def test_noninterface(self):
        """ Check adapt rejects a non-interface. """

        with self.assertRaises(TypeError):
            @adapt(Klass, to=object)
            class NonInterface(Adapter): pass

    def test_issubclass(self):
        """ Check issubclass(). """

        self.assertTrue(issubclass(KlassAdapter, Iface))
        self.assertTrue(issubclass(Klass2Adapter, Iface))
        self.assertTrue(issubclass(Klass2Adapter, Iface2))

    def test_issubclass_fail(self):
        """ Check issubclass() fails. """

        class OtherIface(Interface): pass

        self.assertFalse(issubclass(KlassAdapter, OtherIface))

    def test_isinstance(self):
        """ Check isinstance(). """

        k = KlassAdapter()
        k2 = Klass2Adapter()

        self.assertTrue(isinstance(k, Iface))
        self.assertTrue(isinstance(k2, Iface))
        self.assertTrue(isinstance(k2, Iface2))

    def test_isinstance_fail(self):
        """ Check isinstance() fails. """

        class OtherIface(Interface): pass

        k = KlassAdapter()

        self.assertFalse(isinstance(KlassAdapter, OtherIface))

    def test_adapt(self):
        """ Check the adaptation to an implemented interface. """

        k = Klass()
        iface = Iface(k)

        self.assertEquals(iface.attr, 'iface')
        self.assertIs(iface.adaptee, k)
        self.assertIs(type(iface), KlassAdapter)

    def test_isadaptable(self):
        """ Check adaptation fails when isadaptable() returns False. """

        k = Klass4()

        with self.assertRaises(TypeError):
            Iface(k)

    def test_isadapted_true(self):
        """ Check isadapted() for an adapted object. """

        k = Klass()
        Iface(k)

        self.assertTrue(isadapted(k, Iface))

    def test_isadapted_false(self):
        """ Check isadapted() for an unadapted object. """

        k = Klass()

        self.assertFalse(isadapted(k, Iface))

    def test_adapt_attributes(self):
        """ Check the adaptation of attributes to an implemented interface. """

        class TestModel(Model):
            iface = Instance(Iface)

        k = Klass()
        model = TestModel(iface=k)

        self.assertEquals(Iface(model.iface).attr, 'iface')
        self.assertIs(model.iface, k)

    def test_adapt_singleton(self):
        """ Check the adapters are singletons. """

        k = Klass()
        iface1 = Iface(k)
        iface2 = Iface(k)

        self.assertIs(iface1, iface2)

    def test_adapt_derived(self):
        """ Check using a sub-interface adapter if one is available. """

        k = Klass3()
        iface = Iface(k)

        self.assertIs(type(iface), Klass3Adapter)

    def test_adapt_derived_singleton(self):
        """ Check sub-interface adapters are singletons. """

        k = Klass3()
        iface1 = Iface(k)
        iface2 = SubIface(k)

        self.assertIs(iface1, iface2)

    def test_gc(self):
        """ Check the adaptee can be garbage collected. """

        k = Klass()
        k_ref = weakref.ref(k)
        Iface(k)

        del k
        gc.collect()

        self.assertIsNone(k_ref())

    def test_adapt_interface(self):
        """ Check adapting an interface to an interface. """

        @adapt(Iface, to=Iface2)
        class IfaceToIface2Adapter(Adapter): pass

        @implements(Iface)
        class IfaceImplementation(Model): pass

        impl = IfaceImplementation()
        iface2 = Iface2(impl)

        self.assertEquals(iface2.attr2, 'iface2')

    def test_adapt_fail(self):
        """ Check the adaptation to an unimplemented interface. """

        class OtherIface(Interface): pass

        k = Klass()

        with self.assertRaises(TypeError):
            OtherIface(k)

    def test_duplicate_adapters(self):
        """ Check for duplicate adapters. """

        with self.assertRaises(TypeError):
            @adapt(Klass, to=Iface)
            class DuplicateAdapter(Adapter): pass

    def test_adapter_chain(self):
        """ Check adapters can be chained. """

        class IChain1(Interface): pass
        class IChain2(Interface): pass

        @adapt(Klass, to=IChain1)
        class KlassIChain1Adapter(Adapter):
            pass

        @adapt(IChain1, to=IChain2)
        class IChain1IChain2Adapter(Adapter):
            pass

        k = Klass()

        IChain1(k)

        self.assertIsNotNone(IChain2(k))

    def test_uncached_adapter(self):
        """ Check uncached adapters are discarded. """

        class IUncached(Interface): pass

        @adapt(Klass, to=IUncached)
        class KlassIUncachedAdapter(Adapter):
            pass

        k = Klass()

        self.assertIsNotNone(IUncached(k, cache=False))
        self.assertIsNone(IUncached(k, adapt=False, exception=False))

    def test_explicit_adapter(self):
        """ Test explicit adapters. """

        class IExplicit(Interface): pass

        @implements(IExplicit)
        class KlassIExplicitAdapter(Adapter):
            pass

        k = Klass()

        adapter = KlassIExplicitAdapter(adaptee=k)
        IExplicit(k, adapter=adapter)

        self.assertIs(IExplicit(k), adapter)


if __name__ == '__main__':
    unittest.main()
