# Copyright (c) 2010 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from dip.model import (DelegatedTo, Instance, Model, observe, Str,
        ValidationError)


class ComponentModel(Model):
    """ A test model to be delegated to. """

    value = Str('value')


class Klass(Model):
    """ A test class with a delegate. """

    value = DelegatedTo('component.value')

    observed_value = Str()

    component = Instance(ComponentModel)

    @component.default
    def component(self):
        return ComponentModel()

    @observe('value')
    def _value_changed(self, change):
        self.observed_value = change.new


class TestModel(unittest.TestCase):
    """ This tests the DelegatedTo type. """

    def test_getter(self):
        """ Test the get of a delegated object. """

        k = Klass()

        self.assertEquals(k.value, 'value')

    def test_setter(self):
        """ Test the set of a delegated object. """

        k = Klass()

        k.value = 'new value'
        self.assertEquals(k.component.value, 'new value')

    def test_setter_invalid(self):
        """ Test the set of a delegated object with an invalid value. """

        k = Klass()

        with self.assertRaises(ValidationError):
            k.value = 10

    def test_initial_value(self):
        """ Test the initialisation of a delegated value. """

        k = Klass(value='initial')

        self.assertEquals(k.component.value, 'initial')

    def test_observer(self):
        """ Test the observer for a delegated value. """

        k = Klass()

        k.value = 'observed'
        self.assertEquals(k.observed_value, 'observed')

    def test_observer_delegated(self):
        """ Test the observer when the original value changes. """

        k = Klass()

        k.component.value = 'observed'
        self.assertEquals(k.observed_value, 'observed')

    def test_non_model(self):
        """ Test delegating to a non-model. """

        class ComponentClass:
            def __init__(self, value):
                self.value = value

        class Klass2(Model):

            value = DelegatedTo('component.value')

            component = Instance(ComponentClass)

            @component.default
            def component(self):
                return ComponentClass('initial')

        k = Klass2()

        self.assertEquals(k.value, 'initial')

        k.value = 'new value'
        self.assertEquals(k.component.value, 'new value')


if __name__ == '__main__':
    unittest.main()
