# Copyright (c) 2009 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from dip.model import Instance, Model, ValidationError


class TestInstanceType(unittest.TestCase):
    """ This tests the Instance type. """

    def test_invalid_string_instance_type(self):
        """ Invalid Instance type string. """

        class Klass(Model):
            attr = Instance('foo.bar.baz')

        with self.assertRaises(AttributeError):
            Klass(attr=False)

    def test_valid_string_instance_type(self):
        """ Valid Instance type string. """

        class Klass(Model):
            attr = Instance('dip.model.Model')

        value = Model()
        k = Klass(attr=value)

        self.assertIs(k.attr, value)

    def test_invalid_instance_type(self):
        """ Invalid Instance type specification. """

        class Klass(Model):
            attr = Instance(None)

        with self.assertRaises(TypeError):
            Klass(attr=0)

    def test_default_instance_type(self):
        """ Default Instance type. """

        # Create a valid test class and data.
        class Klass(Model):
            attr = Instance()

        class InvalidClass: pass

        valid_value = Model()
        invalid_value = InvalidClass()

        # Verify the default type allows any instance.
        k = Klass()
        k.attr = valid_value
        self.assertIs(k.attr, valid_value)

        # Verify the default type allows None.
        k = Klass()
        k.attr = None
        self.assertIsNone(k.attr)

    def test_valid_instance_type(self):
        """ Valid Instance type. """

        # Create a valid test class and data.
        class Klass(Model):
            attr = Instance(Model)

        class InvalidClass: pass

        valid_value = Model()
        invalid_value = InvalidClass()

        # Verify a valid value.
        k = Klass(attr=valid_value)
        self.assertIs(k.attr, valid_value)

        # Verify a ``None`` value.
        k = Klass(attr=None)
        self.assertIsNone(k.attr)

        # Verify an invalid value.
        with self.assertRaises(ValidationError):
            Klass(attr=invalid_value)

        # Verify rebinding a valid value.
        k = Klass()
        k.attr = valid_value
        self.assertIs(k.attr, valid_value)

        # Verify rebinding a ``None`` value.
        k = Klass()
        k.attr = None
        self.assertIsNone(k.attr)

        # Verify rebinding an invalid value.
        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr = invalid_value

    def test_multiple_instance_types(self):
        """ Multiple instance types. """

        # Create a valid test class and data.
        class Klass(Model):
            attr = Instance(Model, str)

        valid_value_1 = Model()
        valid_value_2 = 'valid_str'

        # Verify the valid values.
        k = Klass(attr=valid_value_1)
        self.assertIs(k.attr, valid_value_1)

        k = Klass(attr=valid_value_2)
        self.assertEquals(k.attr, valid_value_2)

    def test_explicit_default_instance_type(self):
        """ Instance type with an explicit default. """

        # Create a valid test class and data.
        class Inst(Model): pass

        inst = Inst()

        class Klass(Model):
            attr = Instance(Model, default=inst)

        k = Klass()
        self.assertIs(k.attr, inst)


if __name__ == '__main__':
    unittest.main()
