# Copyright (c) 2011 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from dip.model import implements, Interface, Model, Str


class Iface(Interface):
    """ A test interface. """

    attr = Str()

    attr3 = Str()


class Iface2(Interface):
    """ A second test interface. """

    attr2 = Str('implicit')


class Iface3(Interface):
    """ A third test interface. """

    def foo(self):
        """ This should be reimplemented by an implementation. """


@implements(Iface, Iface2)
class Klass(Model):
    """ A test class that partially implements its interfaces. """

    attr = Str('default')

    attr3 = 'new_value_same_type'


class TestInterface(unittest.TestCase):
    """ This tests interfaces. """

    def test_nonmodel(self):
        """ Check implements rejects a non-model. """

        with self.assertRaises(TypeError):
            @implements(Iface)
            class Klass: pass

    def test_noninterface(self):
        """ Check implements rejects a non-interface. """

        with self.assertRaises(TypeError):
            @implements(object)
            class Klass(Model): pass

    def test_incomplete(self):
        """ Check implements rejects an incomplete implementation. """

        with self.assertRaises(AttributeError):
            @implements(Iface3)
            class Klass(Model): pass

    def test_issubclass(self):
        """ Check issubclass(). """

        self.assertTrue(issubclass(Klass, Iface))
        self.assertTrue(issubclass(Klass, Iface2))

    def test_issubclass(self):
        """ Check multiple implements. """

        @implements(Iface)
        @implements(Iface2)
        class MultiDecorator(Model): pass

        self.assertTrue(issubclass(MultiDecorator, Iface))
        self.assertTrue(issubclass(MultiDecorator, Iface2))

    def test_issubclass_fail(self):
        """ Check issubclass() fails. """

        class OtherIface(Interface): pass

        self.assertFalse(issubclass(Klass, OtherIface))

    def test_isinstance(self):
        """ Check isinstance(). """

        k = Klass()

        self.assertTrue(isinstance(k, Iface))
        self.assertTrue(isinstance(k, Iface2))

    def test_isinstance_fail(self):
        """ Check isinstance() fails. """

        class OtherIface(Interface): pass

        k = Klass()

        self.assertFalse(isinstance(Klass, OtherIface))

    def test_explicit_implementation(self):
        """ Check an explicit implementation. """

        k = Klass()

        self.assertEquals(k.attr, 'default')

    def test_implicit_implementation(self):
        """ Check an implicit implementation. """

        k = Klass()

        self.assertEquals(k.attr2, 'implicit')

    def test_explicit_value(self):
        """ Check an explicit value. """

        k = Klass()

        self.assertEquals(k.attr3, 'new_value_same_type')
        self.assertTrue(isinstance(Klass.attr3, Str))

    def test_cast(self):
        """ Check the casting to an implemented interface. """

        k = Klass()

        self.assertIs(Iface(k), k)

    def test_cast_fail(self):
        """ Check the casting to an unimplemented interface. """

        class OtherIface(Interface): pass

        k = Klass()

        with self.assertRaises(TypeError):
            OtherIface(k)

    def test_cast_fail_no_exception(self):
        """ Check the casting to an unimplemented interface with no exception.
        """

        class OtherIface(Interface): pass

        k = Klass()

        self.assertIsNone(OtherIface(k, exception=False))


if __name__ == '__main__':
    unittest.main()
