# Copyright (c) 2010 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from dip.model import List, Model, observe, Str, ValidationError


class Klass(Model):
    """ A class for test models. """

    default_list = List()

    explicit_default_list = List(default=[None])

    attr = List(Str())

    def __init__(self):
        super().__init__()

        self.change = None
        self.changes = 0
        observe('attr', self, self._observed)

    def _observed(self, change):
        self.change = change
        self.changes += 1


class TestListType(unittest.TestCase):
    """ This tests the List type. """

    def check_observed(self, model, new, old):
        """ Check that a test model observed a change. """

        self.assertEquals(model.changes, 1)

        change = model.change

        self.assertIs(change.model, model)
        self.assertEquals(change.name, 'attr')
        self.assertEquals(change.new, new)
        self.assertEquals(change.old, old)

    def check_unobserved(self, model):
        """ Check that a test model hasn't observed a change. """

        self.assertEquals(model.changes, 0)

    def test_invalid_list_type(self):
        """ Invalid List type. """

        with self.assertRaises(TypeError):
            List(0)

    def test_default_list_type_explicit_default(self):
        """ Default List explicit default value. """

        k = Klass()
        self.assertListEqual(k.explicit_default_list, [None])

    def test_list_subclass_default(self):
        """ Default List sub-class. """

        class SubKlass(Klass):

            default_list = [1]

        k = SubKlass()
        self.assertListEqual(k.default_list, [1])

    def test_default_list_type_default_value(self):
        """ Default List type default value. """

        k = Klass()
        self.assertListEqual(k.default_list, [])

    def test_default_list_type_valid_value(self):
        """ Default List type valid value. """

        k = Klass()
        k.default_list = [1, True, 'one']
        self.assertListEqual(k.default_list, [1, True, 'one'])

    def test_default_list_type_invalid_value(self):
        """ Default List type invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.default_list = None

    def test_list_type_valid_default_value(self):
        """ List type valid default value. """

        k = Klass(attr=['abc', 'def'])
        self.assertListEqual(k.attr, ['abc', 'def'])

    def test_list_type_invalid_default_value(self):
        """ List type invalid default value. """

        with self.assertRaises(ValidationError):
            Klass(attr=[False])

    def test_list_type_rebind_valid_value(self):
        """ List type rebind valid value. """

        k = Klass()
        k.attr = ['abc', 'def']
        self.assertListEqual(k.attr, ['abc', 'def'])
        self.check_observed(k, ['abc', 'def'], [])

    def test_list_type_rebind_unchanged_value(self):
        """ List type rebind unchanged value. """

        k = Klass()
        k.attr = []
        self.assertListEqual(k.attr, [])
        self.check_unobserved(k)

    def test_list_type_rebind_invalid_value(self):
        """ List type rebind invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr = [False]

    def test_list_type_rebind_invalid_list(self):
        """ List type rebind invalid list. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr = k.default_list

    def test_list_append_valid(self):
        """ List append() valid value. """

        k = Klass()
        k.attr.append('abc')
        self.assertListEqual(k.attr, ['abc'])
        self.check_observed(k, ['abc'], [])

    def test_list_append_invalid(self):
        """ List append() invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr.append(False)

    def test_list_extend_valid(self):
        """ List extend() valid value. """

        k = Klass()
        k.attr.extend(['abc'])
        self.assertListEqual(k.attr, ['abc'])
        self.check_observed(k, ['abc'], [])

    def test_list_extend_unchanged(self):
        """ Unchanged list extend(). """

        k = Klass()
        k.attr.extend([])
        self.assertListEqual(k.attr, [])
        self.check_unobserved(k)

    def test_list_extend_invalid(self):
        """ List extend() invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr.extend([False])

    def test_list_insert_valid(self):
        """ List insert() valid value. """

        k = Klass()
        k.attr.insert(0, 'abc')
        self.assertListEqual(k.attr, ['abc'])
        self.check_observed(k, ['abc'], [])

    def test_list_insert_invalid(self):
        """ List insert() invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr.insert(0, False)

    def test_list_pop(self):
        """ List pop(). """

        k = Klass(attr=['abc'])
        element = k.attr.pop()
        self.assertEqual(element, 'abc')
        self.assertListEqual(k.attr, [])
        self.check_observed(k, [], ['abc'])

    def test_list_pop_index(self):
        """ List pop() at an index position. """

        k = Klass(attr=['abc'])
        element = k.attr.pop(0)
        self.assertEqual(element, 'abc')
        self.assertListEqual(k.attr, [])
        self.check_observed(k, [], ['abc'])

    def test_list_remove_valid(self):
        """ List remove() valid value. """

        k = Klass(attr=['abc'])
        k.attr.remove('abc')
        self.assertListEqual(k.attr, [])
        self.check_observed(k, [], ['abc'])

    def test_list_remove_invalid(self):
        """ List remove() invalid value. """

        k = Klass(attr=['abc'])
        with self.assertRaises(ValidationError):
            k.attr.remove(False)

    def test_list_reverse(self):
        """ List reverse(). """

        k = Klass(attr=['xyz', 'abc'])
        k.attr.reverse()
        self.assertListEqual(k.attr, ['abc', 'xyz'])
        self.check_observed(k, ['abc', 'xyz'], ['xyz', 'abc'])

    def test_list_reverse_unchanged(self):
        """ Palendromic list reverse(). """

        k = Klass(attr=['abc', 'abc'])
        k.attr.reverse()
        self.assertListEqual(k.attr, ['abc', 'abc'])
        self.check_unobserved(k)

    def test_list_sort(self):
        """ List sort(). """

        k = Klass(attr=['xyz', 'abc'])
        k.attr.sort()
        self.assertListEqual(k.attr, ['abc', 'xyz'])
        self.check_observed(k, ['abc', 'xyz'], ['xyz', 'abc'])

    def test_list_sort_unchanged(self):
        """ Sorted list sort(). """

        k = Klass(attr=['abc', 'xyz'])
        k.attr.sort()
        self.assertListEqual(k.attr, ['abc', 'xyz'])
        self.check_unobserved(k)

    def test_list_iadd_valid(self):
        """ List += valid value. """

        k = Klass()
        k.attr += ['abc']
        self.assertListEqual(k.attr, ['abc'])
        self.check_observed(k, ['abc'], [])

    def test_list_iadd_unchanged(self):
        """ Unchanged list +=. """

        k = Klass()
        k.attr += []
        self.assertListEqual(k.attr, [])
        self.check_unobserved(k)

    def test_list_iadd_invalid(self):
        """ List += invalid value. """

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr += [False]

    def test_list_indexing_valid(self):
        """ List [] valid value. """

        k = Klass(attr=['xyz'])
        k.attr[0] = 'abc'
        self.assertListEqual(k.attr, ['abc'])
        self.check_observed(k, ['abc'], ['xyz'])

    def test_list_indexing_unchanged(self):
        """ Unchanged list [] valid value. """

        k = Klass(attr=['abc'])
        k.attr[0] = 'abc'
        self.assertListEqual(k.attr, ['abc'])
        self.check_unobserved(k)

    def test_list_indexing_invalid(self):
        """ List [] invalid value. """

        k = Klass(attr=['xyz'])
        with self.assertRaises(ValidationError):
            k.attr[0] = False

    def test_list_slicing_valid(self):
        """ List [:] valid value. """

        k = Klass(attr=['x', 'y', 'z'])
        k.attr[0:2] = ['a', 'b', 'c']
        self.assertListEqual(k.attr, ['a', 'b', 'c', 'z'])
        self.check_observed(k, ['a', 'b', 'c'], ['x', 'y'])

    def test_list_slicing_unchanged(self):
        """ Unchanged List [:]. """

        k = Klass(attr=['x', 'y', 'z'])
        k.attr[0:2] = ['x', 'y']
        self.assertListEqual(k.attr, ['x', 'y', 'z'])
        self.check_unobserved(k)

    def test_list_slicing_invalid(self):
        """ List [:] invalid value. """

        k = Klass(attr=['x', 'y', 'z'])
        with self.assertRaises(ValidationError):
            k.attr[0:2] = [False]

    def test_list_del(self):
        """ List del. """

        k = Klass(attr=['x', 'y', 'z'])
        del k.attr[0]
        self.assertListEqual(k.attr, ['y', 'z'])
        self.check_observed(k, [], ['x'])

    def test_list_del_slice(self):
        """ List del slice. """

        k = Klass(attr=['x', 'y', 'z'])
        del k.attr[0:2]
        self.assertListEqual(k.attr, ['z'])
        self.check_observed(k, [], ['x', 'y'])


if __name__ == '__main__':
    unittest.main()
