# Copyright (c) 2010 Riverbank Computing Limited.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from dip.model import Any, Model, Str, ValidationError


class TestProperties(unittest.TestCase):
    """ This tests implicit properties. """

    def test_implicit_getter(self):
        """ An implicit getter. """

        class Klass(Model):
            @Any()
            def attr(self):
                return 'got'

        k = Klass()

        self.assertEqual(k.attr, 'got')

    def test_getter_setter(self):
        """ A getter and setter. """

        class Klass(Model):
            attr = Any()

            @attr.getter
            def attr(self):
                try:
                    value = self._real_attr
                except AttributeError:
                    value = self._real_attr = 'default'

                return value

            @attr.setter
            def attr(self, value):
                self._real_attr = value.upper()

        # Verify getter and setter.
        k = Klass()

        self.assertEqual(k.attr, 'default')

        k.attr = 'non-default'
        self.assertEqual(k.attr, 'NON-DEFAULT')

    def test_getter_setter_args(self):
        """ A getter and setter passed as arguments to the type. """

        class Klass(Model):
            attr = Any(getter=lambda s: s.get_real(), setter=lambda s, v: s.set_real(v))

            def get_real(self):
                return self.real

            def set_real(self, value):
                self.real = value

        k = Klass()

        k.attr = 'first'
        self.assertEqual(k.real, 'first')

        k.real = 'second'
        self.assertEqual(k.attr, 'second')

    def test_default(self):
        """ Default values. """

        class Klass(Model):
            attr = Any()

            @attr.default
            def attr(self):
                return 'default'

            @attr.getter
            def attr(self):
                return self._attr

            @attr.setter
            def attr(self, value):
                self._attr = value

        # Verify the default handler is called exactly once.
        k = Klass()
        self.assertEqual(k.attr, 'default')
        k.attr = 'non-default'
        self.assertEqual(k.attr, 'non-default')

        # Verify the default handler is per instance, not per type.
        k = Klass()
        self.assertEqual(k.attr, 'default')

        # Verify the default handler is overridden.
        k = Klass(attr='__init__')
        self.assertEqual(k.attr, '__init__')
        k.attr = 'non-default'
        self.assertEqual(k.attr, 'non-default')

    def test_subclass_default(self):
        """ Sub-class value overrides decorated default. """

        class Klass(Model):
            attr = Any()

            @attr.default
            def attr(self):
                return 'Klass default'

        class Subklass(Klass):
            attr = 'Subklass default'

        # Verify the sub-class value overrides the computed super-class
        # default.
        k = Subklass()
        self.assertEqual(k.attr, 'Subklass default')

    def test_validate_get(self):
        """ Validate getting of values. """

        class Klass(Model):
            attr = Str()

            @attr.getter
            def attr(self):
                return False

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr

    def test_validate_set(self):
        """ Validate setting of values. """

        class Klass(Model):
            attr = Str()

            @attr.getter
            def attr(self):
                return 'valid'

            @attr.setter
            def attr(self, value):
                pass

        k = Klass()
        with self.assertRaises(ValidationError):
            k.attr = False

        with self.assertRaises(ValidationError):
            Klass(attr=False)


if __name__ == '__main__':
    unittest.main()
